part of katana_form;

/// Widgets for text fields for forms.
///
/// Place [FormController.key] under the given [Form].
///
/// Enter the initial value given by [FormController.value] in [initialValue].
///
/// Each time the content is changed, [onChanged] is executed.
///
/// When [FormController.validateAndSave] is executed, validation and data saving are performed.
///
/// Only when [emptyErrorText] is specified, [emptyErrorText] will be displayed as an error if no characters are entered.
/// Only when [lengthErrorText] is specified, if the number of characters entered is less than [minLength], it is displayed as [lengthErrorText].
///
/// Other error checking is performed by specifying [validator].
/// If a string other than [Null] is returned in the callback, the string is displayed as an error statement. If [Null] is returned, it is processed as no error.
///
/// The contents of [onSaved] are executed to save the data.
///
/// The [onSubmitted] process is executed when the Enter key or other keys are pressed.
///
/// If [suggestion] is specified, suggestions will be displayed according to what you have entered.
///
/// If [onDeleteSuggestion] is specified, a delete button is displayed on the displayed suggestions, and if the button is pressed, the process is executed when the suggestion is deleted.
///
/// If [enabled] is `false`, the text is deactivated.
///
/// If [readOnly] is set to `true`, the activation is displayed, but the text cannot be changed.
///
/// If [obscureText] is set to `true`, the input string will be hidden. Please use this function for inputting passwords, etc.
///
/// If [inputFormatters] is specified, it is possible to restrict the characters to be entered.
///
/// フォーム用のテキストフィールド用のウィジェット。
///
/// [FormController.key]を与えた[Form]配下に配置します。
///
/// [initialValue]に[FormController.value]から与えられた初期値を入力します。
///
/// 内容が変更される度[onChanged]が実行されます。
///
/// [FormController.validateAndSave]が実行された場合、バリデーションとデータの保存を行ないます。
///
/// [emptyErrorText]が指定されている時に限り、文字が入力されていない場合[emptyErrorText]がエラーとして表示されます。
/// [lengthErrorText]が指定されている時に限り、[minLength]より入力された文字数が少ない場合[lengthErrorText]として表示されます。
///
/// それ以外のエラーチェックは[validator]を指定することで行ないます。
/// コールバック内で[Null]以外を返すようにするとその文字列がエラー文として表示されます。[Null]の場合はエラーなしとして処理されます。
///
/// データの保存は[onSaved]の内容が実行されます。
///
/// Enterキーなどが押された場合の処理を[onSubmitted]が実行されます。
///
/// [suggestion]が指定されている場合、入力した内容に応じてサジェストが表示されます。
///
/// [onDeleteSuggestion]が指定されている場合、表示されたサジェストに削除ボタンが表示され、ボタンが押された場合、サジェスト削除時の処理が実行されます。
///
/// [enabled]が`false`になるとテキストが非有効化されます。
///
/// [readOnly]が`true`になっている場合は、有効化の表示になりますが、テキストが変更できなくなります。
///
/// [obscureText]が`true`になると入力された文字列が隠されます。パスワードの入力などにご利用ください。
///
/// [inputFormatters]が指定されると入力される文字を制限することが可能です。
class FormItemTextField extends StatefulWidget {
  /// Widgets for text fields for forms.
  ///
  /// Place [FormController.key] under the given [Form].
  ///
  /// Enter the initial value given by [FormController.value] in [initialValue].
  ///
  /// Each time the content is changed, [onChanged] is executed.
  ///
  /// When [FormController.validateAndSave] is executed, validation and data saving are performed.
  ///
  /// Only when [emptyErrorText] is specified, [emptyErrorText] will be displayed as an error if no characters are entered.
  /// Only when [lengthErrorText] is specified, if the number of characters entered is less than [minLength], it is displayed as [lengthErrorText].
  ///
  /// Other error checking is performed by specifying [validator].
  /// If a string other than [Null] is returned in the callback, the string is displayed as an error statement. If [Null] is returned, it is processed as no error.
  ///
  /// The contents of [onSaved] are executed to save the data.
  ///
  /// The [onSubmitted] process is executed when the Enter key or other keys are pressed.
  ///
  /// If [suggestion] is specified, suggestions will be displayed according to what you have entered.
  ///
  /// If [onDeleteSuggestion] is specified, a delete button is displayed on the displayed suggestions, and if the button is pressed, the process is executed when the suggestion is deleted.
  ///
  /// If [enabled] is `false`, the text is deactivated.
  ///
  /// If [readOnly] is set to `true`, the activation is displayed, but the text cannot be changed.
  ///
  /// If [obscureText] is set to `true`, the input string will be hidden. Please use this function for inputting passwords, etc.
  ///
  /// If [inputFormatters] is specified, it is possible to restrict the characters to be entered.
  ///
  /// フォーム用のテキストフィールド用のウィジェット。
  ///
  /// [FormController.key]を与えた[Form]配下に配置します。
  ///
  /// [initialValue]に[FormController.value]から与えられた初期値を入力します。
  ///
  /// 内容が変更される度[onChanged]が実行されます。
  ///
  /// [FormController.validateAndSave]が実行された場合、バリデーションとデータの保存を行ないます。
  ///
  /// [emptyErrorText]が指定されている時に限り、文字が入力されていない場合[emptyErrorText]がエラーとして表示されます。
  /// [lengthErrorText]が指定されている時に限り、[minLength]より入力された文字数が少ない場合[lengthErrorText]として表示されます。
  ///
  /// それ以外のエラーチェックは[validator]を指定することで行ないます。
  /// コールバック内で[Null]以外を返すようにするとその文字列がエラー文として表示されます。[Null]の場合はエラーなしとして処理されます。
  ///
  /// データの保存は[onSaved]の内容が実行されます。
  ///
  /// Enterキーなどが押された場合の処理を[onSubmitted]が実行されます。
  ///
  /// [suggestion]が指定されている場合、入力した内容に応じてサジェストが表示されます。
  ///
  /// [onDeleteSuggestion]が指定されている場合、表示されたサジェストに削除ボタンが表示され、ボタンが押された場合、サジェスト削除時の処理が実行されます。
  ///
  /// [enabled]が`false`になるとテキストが非有効化されます。
  ///
  /// [readOnly]が`true`になっている場合は、有効化の表示になりますが、テキストが変更できなくなります。
  ///
  /// [obscureText]が`true`になると入力された文字列が隠されます。パスワードの入力などにご利用ください。
  ///
  /// [inputFormatters]が指定されると入力される文字を制限することが可能です。
  const FormItemTextField({
    super.key,
    this.controller,
    this.style,
    this.keyboardType = TextInputType.text,
    this.maxLength,
    this.onTap,
    this.minLength,
    this.maxLines,
    this.minLines = 1,
    this.expands = false,
    this.hintText,
    this.labelText,
    this.lengthErrorText,
    this.prefix,
    this.suffix,
    this.prefixText,
    this.suffixText,
    this.prefixIcon,
    this.prefixIconConstraints,
    this.suffixIcon,
    this.suffixIconConstraints,
    this.suggestion = const [],
    this.enabled = true,
    this.readOnly = false,
    this.obscureText = false,
    this.counterText = "",
    this.onDeleteSuggestion,
    this.validator,
    this.inputFormatters,
    this.onSaved,
    this.onSubmitted,
    this.onChanged,
    this.showCursor,
    this.focusNode,
    this.emptyErrorText,
    this.initialValue,
  });

  /// Form Style.
  ///
  /// フォームのスタイル。
  final FormStyle? style;

  /// Controller for text forms.
  ///
  /// テキストフォーム用のコントローラー。
  final TextEditingController? controller;

  /// Mobile software keyboard type.
  ///
  /// モバイルのソフトウェアキーボードタイプ。
  final TextInputType keyboardType;

  /// Initial value.
  ///
  /// 初期値。
  final String? initialValue;

  /// Maximum number of strings. You will not be able to enter more than this.
  ///
  /// 文字列の最大数。これ以上は入力できなくなります。
  final int? maxLength;

  /// Minimum number of strings. If [lengthErrorText] is specified, an error is displayed.
  ///
  /// 文字列の最小数。[lengthErrorText]が指定されている場合、エラーが表示されます。
  final int? minLength;

  /// Maximum number of lines. No more line breaks will be allowed.
  ///
  /// 最大のライン数。これ以上は改行できなくなります。
  final int? maxLines;

  /// Minimum number of lines; if a number greater than 1 is specified, the initial display of the form extends to the specified number of lines.
  ///
  /// 最小のライン数。1より大きい数が指定されている場合、フォームの初期表示が指定された行数分に広がります。
  final int minLines;

  /// Hint to be displayed on the form. Displayed when no text is entered.
  ///
  /// フォームに表示するヒント。文字が入力されていない場合表示されます。
  final String? hintText;

  /// Error text. Only displayed if no characters are entered.
  ///
  /// エラーテキスト。入力された文字がない場合のみ表示されます。
  final String? emptyErrorText;

  /// Label text for forms.
  ///
  /// フォーム用のラベルテキスト。
  final String? labelText;

  /// Text of the character counter. Default is disabled.
  ///
  /// 文字数のカウンターのテキスト。デフォルトは無効化されています。
  final String? counterText;

  /// Text to be displayed if the number of characters is less than or equal to [minLength]. If this is not [Null], an error is checked and displayed.
  ///
  /// 文字数が[minLength]以下の場合に表示するテキスト。これが[Null]以外の場合、エラーチェックされ表示されます。
  final String? lengthErrorText;

  /// If this is `false`, it is deactivated.
  ///
  /// In addition to disabling input, the form design, etc., will also be changed to a deactivated version.
  ///
  /// これが`false`の場合、非有効化されます。
  ///
  /// 入力ができなくなる他、フォームのデザイン等も非有効化されたものに変更されます。
  final bool enabled;

  /// フォームの文字の左側に表示されるウィジェット。
  final Widget? prefix;
  final Widget? suffix;
  final Widget? prefixIcon;
  final Widget? suffixIcon;
  final String? prefixText;
  final String? suffixText;
  final BoxConstraints? prefixIconConstraints;
  final BoxConstraints? suffixIconConstraints;
  final bool readOnly;
  final bool obscureText;
  final bool expands;
  final List<String> suggestion;
  final void Function(String value)? onDeleteSuggestion;
  final void Function(String? value)? onSaved;
  final void Function(String? value)? onChanged;
  final String? Function(String? value)? validator;
  final bool? showCursor;
  final VoidCallback? onTap;
  final FocusNode? focusNode;
  final List<TextInputFormatter>? inputFormatters;
  final void Function(String? value)? onSubmitted;

  @override
  State<StatefulWidget> createState() => _FormItemTextFieldState();
}

class _FormItemTextFieldState extends State<FormItemTextField> {
  TextEditingController? get _effectiveController =>
      widget.controller ?? _controller;
  TextEditingController? _controller;

  @override
  void initState() {
    super.initState();
    if (widget.controller != null) {
      _controller = TextEditingController(text: widget.initialValue);
    }
    _effectiveController?.addListener(_handledOnUpdate);
  }

  @override
  void didUpdateWidget(FormItemTextField oldWidget) {
    super.didUpdateWidget(oldWidget);

    if (oldWidget.controller != widget.controller) {
      oldWidget.controller?.removeListener(_handledOnUpdate);
      widget.controller?.addListener(_handledOnUpdate);
    }
    if (oldWidget.initialValue != widget.initialValue &&
        widget.initialValue != null) {
      _effectiveController?.text = widget.initialValue!;
    }
  }

  @override
  void dispose() {
    super.dispose();
    _effectiveController?.removeListener(_handledOnUpdate);
    _controller?.dispose();
  }

  void _handledOnUpdate() {
    setState(() {});
  }

  @override
  Widget build(BuildContext context) {
    final mainTextStyle = widget.style?.textStyle?.copyWith(
          color: widget.style?.color,
        ) ??
        TextStyle(
          color:
              widget.style?.color ?? Theme.of(context).colorScheme.onBackground,
        );
    final subTextStyle = widget.style?.textStyle?.copyWith(
          color: widget.style?.subColor,
        ) ??
        TextStyle(
          color: widget.style?.subColor ??
              widget.style?.color?.withOpacity(0.5) ??
              Theme.of(context).colorScheme.onBackground.withOpacity(0.5),
        );
    final errorTextStyle = widget.style?.textStyle?.copyWith(
          color: widget.style?.errorColor,
        ) ??
        TextStyle(
          color:
              widget.style?.errorColor ?? Theme.of(context).colorScheme.error,
        );
    const borderSide = OutlineInputBorder(
      borderSide: BorderSide.none,
    );

    return SuggestionOverlayBuilder(
      items: widget.suggestion,
      onDeleteSuggestion: widget.onDeleteSuggestion,
      controller: widget.controller,
      focusNode: widget.focusNode,
      builder: (context, controller, onTap) => Container(
        height: widget.style?.height,
        width: widget.style?.width,
        padding:
            widget.style?.padding ?? const EdgeInsets.symmetric(vertical: 16),
        child: TextFormField(
          key: widget.key,
          cursorColor: widget.style?.cursorColor,
          inputFormatters: widget.inputFormatters,
          focusNode: widget.focusNode,
          textAlign: widget.style?.textAlign ?? TextAlign.left,
          textAlignVertical: widget.style?.textAlignVertical,
          showCursor: widget.showCursor,
          enabled: widget.enabled,
          controller: controller,
          keyboardType: widget.keyboardType,
          maxLength: widget.maxLength,
          maxLines: widget.obscureText
              ? 1
              : (widget.expands ? null : widget.maxLines),
          minLines: widget.obscureText
              ? 1
              : (widget.expands ? null : widget.minLines),
          expands: !widget.obscureText && widget.expands,
          decoration: InputDecoration(
            contentPadding: widget.style?.contentPadding ??
                const EdgeInsets.symmetric(horizontal: 16, vertical: 0),
            fillColor: widget.style?.backgroundColor,
            filled: widget.style?.backgroundColor != null,
            isDense: true,
            border: widget.style?.border ?? borderSide,
            enabledBorder: widget.style?.border ?? borderSide,
            disabledBorder: widget.style?.disabledBorder ??
                widget.style?.border ??
                borderSide,
            errorBorder:
                widget.style?.errorBorder ?? widget.style?.border ?? borderSide,
            focusedBorder: widget.style?.border ?? borderSide,
            focusedErrorBorder:
                widget.style?.errorBorder ?? widget.style?.border ?? borderSide,
            hintText: widget.hintText,
            labelText: widget.labelText,
            counterText: widget.counterText,
            prefix: widget.prefix,
            suffix: widget.suffix,
            prefixIcon: widget.prefixIcon,
            suffixIcon: widget.suffixIcon,
            prefixText: widget.prefixText,
            suffixText: widget.suffixText,
            prefixIconConstraints: widget.prefixIconConstraints,
            suffixIconConstraints: widget.suffixIconConstraints,
            labelStyle: mainTextStyle,
            hintStyle: subTextStyle,
            suffixStyle: subTextStyle,
            prefixStyle: subTextStyle,
            counterStyle: subTextStyle,
            helperStyle: subTextStyle,
            errorStyle: errorTextStyle,
          ),
          style: mainTextStyle,
          obscureText: widget.obscureText,
          readOnly: widget.readOnly,
          onFieldSubmitted: (value) {
            widget.onSubmitted?.call(value);
          },
          onTap: widget.enabled
              ? () {
                  if (widget.onTap != null) {
                    widget.onTap?.call();
                  } else {
                    onTap.call();
                  }
                }
              : null,
          validator: (value) {
            if (widget.emptyErrorText.isNotEmpty && value.isEmpty) {
              return widget.emptyErrorText;
            }
            if (widget.lengthErrorText.isNotEmpty &&
                widget.minLength.def(0) > value.length) {
              return widget.lengthErrorText;
            }
            return widget.validator?.call(value);
          },
          onChanged: widget.onChanged,
          onSaved: (value) {
            widget.onSaved?.call(value);
          },
        ),
      ),
    );
  }
}

class SuggestionOverlayBuilder extends StatefulWidget {
  const SuggestionOverlayBuilder({
    super.key,
    required this.builder,
    required this.items,
    this.onDeleteSuggestion,
    this.maxHeight = 260,
    this.color,
    this.focusNode,
    this.backgroundColor,
    this.elevation = 8.0,
    this.controller,
    this.onTap,
    this.offset = const Offset(0, 20),
    this.showOnTap = true,
  });

  final List<String> items;

  final FocusNode? focusNode;

  final Offset offset;

  final TextEditingController? controller;

  final VoidCallback? onTap;

  final double elevation;

  final Color? backgroundColor;

  final void Function(String value)? onDeleteSuggestion;

  final Color? color;

  final Widget Function(
    BuildContext context,
    TextEditingController controller,
    VoidCallback onTap,
  ) builder;

  final bool showOnTap;

  final double maxHeight;

  @override
  State<StatefulWidget> createState() => _SuggestionOverlayBuilderState();
}

class _SuggestionOverlayBuilderState extends State<SuggestionOverlayBuilder> {
  OverlayEntry? _overlay;
  TextEditingController? _controller;
  TextEditingController? get _effectiveController =>
      widget.controller ?? _controller;

  final LayerLink _layerLink = LayerLink();
  @override
  void initState() {
    super.initState();
    _controller = widget.controller ?? TextEditingController();
    widget.focusNode?.addListener(_listener);
    _effectiveController?.addListener(_listener);
  }

  @override
  void didUpdateWidget(SuggestionOverlayBuilder oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.controller != oldWidget.controller) {
      oldWidget.controller?.removeListener(_listener);
      widget.controller?.addListener(_listener);

      if (oldWidget.controller != null && widget.controller == null)
        _controller =
            TextEditingController.fromValue(oldWidget.controller?.value);
      if (widget.controller != null) {
        if (oldWidget.controller == null) {
          _controller = null;
        }
      }
    }
    if (widget.focusNode != oldWidget.focusNode) {
      oldWidget.focusNode?.removeListener(_listener);
      widget.focusNode?.addListener(_listener);
    }
  }

  void _listener() {
    if (_overlay != null) {
      return;
    }
    if (_effectiveController == null) {
      return;
    }
    if (widget.items.isEmpty) {
      return;
    }
    final search = _effectiveController?.text;
    final wordList = search?.split(" ") ?? const <String>[];
    if (widget.focusNode != null &&
        widget.focusNode!.hasFocus &&
        search.isEmpty) {
      _updateOverlay();
      return;
    }
    if (!widget.items.any(
      (element) =>
          element.isNotEmpty &&
          wordList.isNotEmpty &&
          wordList.last.isNotEmpty &&
          element != wordList.last &&
          element.toLowerCase().startsWith(wordList.last.toLowerCase()),
    )) {
      return;
    }
    _updateOverlay();
  }

  @override
  void dispose() {
    super.dispose();
    widget.focusNode?.removeListener(_listener);
    _effectiveController?.removeListener(_listener);
  }

  @override
  Widget build(BuildContext context) {
    if (widget.items.isEmpty) {
      return widget.builder(
        context,
        _effectiveController!,
        widget.onTap ?? () {},
      );
    }
    return WillPopScope(
      onWillPop: () {
        if (_overlay == null) {
          return Future.value(true);
        } else {
          _overlay!.remove();
          _overlay = null;
          return Future.value(false);
        }
      },
      child: CompositedTransformTarget(
        link: _layerLink,
        child: widget.builder(
          context,
          _effectiveController!,
          () {
            if (widget.showOnTap) {
              _updateOverlay();
            }
            widget.onTap?.call();
          },
        ),
      ),
    );
  }

  void _updateOverlay() {
    final itemBox = context.findRenderObject() as RenderBox?;
    if (itemBox == null) {
      return;
    }
    final textFieldSize = itemBox.size;
    final width = textFieldSize.width;
    final height = textFieldSize.height;
    final rect = itemBox.localToGlobal(Offset.zero) & textFieldSize;
    final screen = MediaQuery.of(context).size;
    final up = rect.top > (screen.height / 2.0);
    _overlay = OverlayEntry(
      builder: (context) => Stack(
        children: [
          GestureDetector(
            onTap: () {
              _overlay?.remove();
              _overlay = null;
            },
            child: Container(
              constraints: const BoxConstraints.expand(),
              color: Colors.transparent,
            ),
          ),
          Positioned(
            width: width,
            child: CompositedTransformFollower(
              link: _layerLink,
              showWhenUnlinked: false,
              offset: Offset(0.0, -widget.maxHeight),
              child: SizedBox(
                width: width,
                child: _SuggestionOverlay(
                  items: widget.items,
                  color:
                      widget.color ?? Theme.of(context).colorScheme.onSurface,
                  offset: Offset(
                    widget.offset.dx,
                    up
                        ? widget.offset.dy
                        : widget.maxHeight + height + widget.offset.dy,
                  ),
                  maxHeight: widget.maxHeight,
                  direction: up ? VerticalDirection.up : VerticalDirection.down,
                  onDeleteSuggestion: widget.onDeleteSuggestion,
                  backgroundColor: widget.backgroundColor ??
                      Theme.of(context).colorScheme.surface,
                  controller: _effectiveController!,
                  elevation: widget.elevation,
                  onTap: () {
                    _overlay?.remove();
                    _overlay = null;
                  },
                ),
              ),
            ),
          ),
        ],
      ),
    );
    if (_overlay != null) {
      Navigator.of(context).overlay?.insert(_overlay!);
    }
  }
}

class _SuggestionOverlay extends StatefulWidget {
  const _SuggestionOverlay({
    required this.items,
    this.controller,
    this.offset = Offset.zero,
    this.maxHeight = 260,
    this.direction = VerticalDirection.down,
    this.onDeleteSuggestion,
    this.elevation = 8.0,
    this.color = Colors.black,
    this.onTap,
    this.backgroundColor = Colors.white,
  });
  final double elevation;
  final Color backgroundColor;
  final Color color;
  final List<String> items;
  final VerticalDirection direction;
  final void Function(String value)? onDeleteSuggestion;
  final TextEditingController? controller;
  final VoidCallback? onTap;
  final double maxHeight;
  final Offset offset;
  @override
  State<StatefulWidget> createState() => _SuggestionOverlayState();
}

class _SuggestionOverlayState extends State<_SuggestionOverlay> {
  String? _search;
  TextEditingController? _controller;
  TextEditingController? get _effectiveController =>
      widget.controller ?? _controller;

  List<String> _wordList = [];
  @override
  void initState() {
    super.initState();
    _controller = widget.controller ?? TextEditingController();
    _search = _effectiveController?.text;
    _wordList = _search?.split(" ") ?? const <String>[];
    _effectiveController?.addListener(_listener);
  }

  @override
  void didUpdateWidget(_SuggestionOverlay oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.controller != oldWidget.controller) {
      oldWidget.controller?.removeListener(_listener);
      widget.controller?.addListener(_listener);

      if (oldWidget.controller != null && widget.controller == null)
        _controller =
            TextEditingController.fromValue(oldWidget.controller?.value);
      if (widget.controller != null) {
        if (oldWidget.controller == null) {
          _controller = null;
        }
      }
    }
  }

  void _listener() {
    setState(() {
      _search = _effectiveController?.text;
      _wordList = _search?.split(" ") ?? const <String>[];
    });
  }

  @override
  void dispose() {
    super.dispose();
    _effectiveController?.removeListener(_listener);
  }

  @override
  Widget build(BuildContext context) {
    final widgets = widget.items.mapAndRemoveEmpty(
      (e) {
        if (e.isNotEmpty &&
            _wordList.isNotEmpty &&
            _wordList.last.isNotEmpty &&
            e != _wordList.last &&
            !e.toLowerCase().startsWith(_wordList.last.toLowerCase())) {
          return null;
        }
        return GestureDetector(
          onTap: () {
            if (_wordList.isNotEmpty) {
              _wordList[_wordList.length - 1] = e;
            }
            final text = _wordList.join(" ");
            _effectiveController?.clearComposing();
            _effectiveController?.clear();
            _effectiveController?.text = text;
            _effectiveController?.selection = TextSelection.fromPosition(
              TextPosition(offset: _effectiveController?.text.length ?? 0),
            );
            widget.onTap?.call();
          },
          child: Container(
            padding: const EdgeInsets.all(10),
            constraints: const BoxConstraints.expand(height: 50),
            child: Row(
              children: [
                Expanded(
                  child: Text(
                    e,
                    style: TextStyle(fontSize: 18, color: widget.color),
                  ),
                ),
                if (widget.onDeleteSuggestion != null)
                  Container(
                    width: 80,
                    alignment: Alignment.centerRight,
                    child: IconButton(
                      visualDensity: VisualDensity.compact,
                      padding: const EdgeInsets.all(0),
                      icon: Icon(Icons.close, size: 20, color: widget.color),
                      onPressed: () {
                        setState(() {
                          widget.items.remove(e);
                          widget.onDeleteSuggestion?.call(e);
                        });
                      },
                    ),
                  ),
              ],
            ),
          ),
        );
      },
    );
    if (widgets.isEmpty) {
      widget.onTap?.call();
      return const SizedBox.shrink();
    }
    final height = min((widgets.length * 50).toDouble() + 20, widget.maxHeight);

    final offset = widget.offset.dy +
        (widget.direction == VerticalDirection.down
            ? 0
            : (widget.maxHeight - height));
    return Container(
      height: height + offset,
      padding: EdgeInsets.only(top: offset),
      child: Card(
        elevation: widget.elevation,
        shape: const RoundedRectangleBorder(borderRadius: BorderRadius.zero),
        color: widget.backgroundColor,
        child: SingleChildScrollView(
          reverse: widget.direction == VerticalDirection.up,
          padding: const EdgeInsets.symmetric(horizontal: 10, vertical: 5),
          child: Column(
            verticalDirection: widget.direction,
            crossAxisAlignment: CrossAxisAlignment.start,
            mainAxisSize: MainAxisSize.min,
            children: widgets,
          ),
        ),
      ),
    );
  }
}
