part of katana_form;

class FormNumField extends StatefulWidget {
  const FormNumField({
    super.key,
    this.controller,
    this.keyboardType = TextInputType.number,
    this.hintText,
    this.labelText,
    this.style,
    this.enabled = true,
    this.prefix,
    this.suffix,
    this.prefixText,
    this.suffixText,
    this.prefixIcon,
    this.prefixIconConstraints,
    this.suffixIcon,
    this.suffixIconConstraints,
    this.errorText,
    this.allowEmpty = false,
    this.readOnly = false,
    this.obscureText = false,
    this.validator,
    this.textAlign = TextAlign.start,
    this.onChanged,
    this.onEditingComplete,
    this.contentPadding,
    this.initialValue,
    this.picker = const FormNumFieldPicker(),
    this.onSaved,
  });

  final FormStyle? style;
  final TextEditingController? controller;
  final TextInputType keyboardType;
  final String? hintText;
  final String? errorText;
  final String? labelText;
  final TextAlign textAlign;
  final bool readOnly;
  final bool allowEmpty;
  final bool obscureText;
  final num? initialValue;
  final bool enabled;
  final Widget? prefix;
  final Widget? suffix;
  final Widget? prefixIcon;
  final Widget? suffixIcon;
  final String? prefixText;
  final String? suffixText;
  final BoxConstraints? prefixIconConstraints;
  final BoxConstraints? suffixIconConstraints;
  final FormFieldValidator<num?>? validator;
  final EdgeInsetsGeometry? contentPadding;
  final void Function(num? value)? onSaved;
  final void Function(num? value)? onChanged;
  final void Function()? onEditingComplete;
  final FormNumFieldPicker picker;

  @override
  State<StatefulWidget> createState() => _FormNumFieldState();
}

class _FormNumFieldState extends State<FormNumField> {
  TextEditingController? _controller;
  @override
  void initState() {
    super.initState();
    if (widget.controller == null) {
      return;
    }
    if (widget.initialValue != null) {
      widget.controller?.text = widget.initialValue!.toString();
    }
    _controller = TextEditingController(text: widget.controller?.text);
    _controller?.addListener(_listenerInside);
    widget.controller?.addListener(_listenerOutside);
  }

  @override
  void dispose() {
    super.dispose();
    _controller?.removeListener(_listenerInside);
    widget.controller?.removeListener(_listenerOutside);
  }

  void _listenerOutside() {
    if (_controller?.text == widget.controller?.text) {
      return;
    }
    _controller?.text = widget.controller?.text ?? "";
  }

  void _listenerInside() {
    if (_controller?.text == widget.controller?.text) {
      return;
    }
    widget.controller?.text = _controller?.text ?? "";
  }

  @override
  Widget build(BuildContext context) {
    final mainTextStyle = widget.style?.textStyle?.copyWith(
          color: widget.style?.color,
        ) ??
        TextStyle(
          color:
              widget.style?.color ?? Theme.of(context).colorScheme.onBackground,
        );
    final subTextStyle = widget.style?.textStyle?.copyWith(
          color: widget.style?.subColor,
        ) ??
        TextStyle(
          color: widget.style?.subColor ??
              widget.style?.color?.withOpacity(0.5) ??
              Theme.of(context).colorScheme.onBackground.withOpacity(0.5),
        );
    final errorTextStyle = widget.style?.textStyle?.copyWith(
          color: widget.style?.errorColor,
        ) ??
        TextStyle(
          color:
              widget.style?.errorColor ?? Theme.of(context).colorScheme.error,
        );
    const borderSide = OutlineInputBorder(
      borderSide: BorderSide.none,
    );

    return Padding(
      padding:
          widget.style?.padding ?? const EdgeInsets.symmetric(vertical: 16),
      child: _NumTextField(
        controller: _controller,
        keyboardType: TextInputType.text,
        initialValue: widget.initialValue,
        enabled: widget.enabled,
        decoration: InputDecoration(
          contentPadding: widget.style?.contentPadding ??
              const EdgeInsets.symmetric(horizontal: 16, vertical: 0),
          fillColor: widget.style?.backgroundColor,
          filled: widget.style?.backgroundColor != null,
          isDense: true,
          border: widget.style?.border ?? borderSide,
          enabledBorder: widget.style?.border ?? borderSide,
          disabledBorder: widget.style?.disabledBorder ??
              widget.style?.border ??
              borderSide,
          errorBorder:
              widget.style?.errorBorder ?? widget.style?.border ?? borderSide,
          focusedBorder: widget.style?.border ?? borderSide,
          focusedErrorBorder:
              widget.style?.errorBorder ?? widget.style?.border ?? borderSide,
          hintText: widget.hintText,
          labelText: widget.labelText,
          prefix: widget.prefix,
          suffix: widget.suffix,
          prefixIcon: widget.prefixIcon,
          suffixIcon: widget.suffixIcon,
          prefixText: widget.prefixText,
          suffixText: widget.suffixText,
          prefixIconConstraints: widget.prefixIconConstraints,
          suffixIconConstraints: widget.suffixIconConstraints,
          labelStyle: mainTextStyle,
          hintStyle: subTextStyle,
          suffixStyle: subTextStyle,
          prefixStyle: subTextStyle,
          counterStyle: subTextStyle,
          helperStyle: subTextStyle,
          errorStyle: errorTextStyle,
        ),
        style: mainTextStyle,
        obscureText: widget.obscureText,
        textAlign: widget.style?.textAlign ?? TextAlign.left,
        textAlignVertical: widget.style?.textAlignVertical,
        readOnly: widget.readOnly,
        validator: (value) {
          if (!widget.allowEmpty &&
              widget.errorText.isNotEmpty &&
              value == null) {
            return widget.errorText;
          }
          return widget.validator?.call(value);
        },
        onEditingComplete: widget.onEditingComplete,
        onChanged: widget.onChanged,
        onSaved: (value) {
          if (!widget.allowEmpty && value == null) {
            return;
          }
          widget.onSaved?.call(value);
        },
        picker: widget.picker,
      ),
    );
  }
}

class _NumTextField extends FormField<num> {
  _NumTextField({
    required this.picker,
    Key? key,
    FormFieldSetter<num?>? onSaved,
    FormFieldValidator<num?>? validator,
    num? initialValue,
    bool enabled = true,
    this.onChanged,
    this.controller,
    // ignore: unused_element
    this.focusNode,
    InputDecoration decoration = const InputDecoration(),
    TextInputType? keyboardType,
    TextCapitalization textCapitalization = TextCapitalization.none,
    TextInputAction? textInputAction,
    TextStyle? style,
    StrutStyle? strutStyle,
    TextAlign textAlign = TextAlign.start,
    TextAlignVertical? textAlignVertical,
    bool autofocus = false,
    this.readOnly = true,
    bool? showCursor,
    bool obscureText = false,
    bool autocorrect = true,
    bool expands = false,
    VoidCallback? onEditingComplete,
    ValueChanged<num?>? onSubmitted,
    List<TextInputFormatter>? inputFormatters,
    double cursorWidth = 2.0,
    Radius? cursorRadius,
    Color? cursorColor,
    Brightness? keyboardAppearance,
    EdgeInsets scrollPadding = const EdgeInsets.all(20.0),
    bool enableInteractiveSelection = true,
    InputCounterWidgetBuilder? buildCounter,
  }) : super(
          key: key,
          initialValue: initialValue,
          enabled: enabled,
          validator: validator,
          onSaved: onSaved,
          builder: (field) {
            final _NumTextFieldState state = field as _NumTextFieldState;
            final InputDecoration effectiveDecoration = decoration
                .applyDefaults(Theme.of(field.context).inputDecorationTheme);
            return TextField(
              mouseCursor: SystemMouseCursors.click,
              controller: state._effectiveController ??
                  TextEditingController(
                    text: state.value != null ? state.value!.toString() : null,
                  ),
              focusNode: state._effectiveFocusNode,
              decoration: effectiveDecoration.copyWith(
                errorText: field.errorText,
              ),
              keyboardType: keyboardType,
              textAlign: textAlign,
              textAlignVertical: textAlignVertical,
              textInputAction: textInputAction,
              style: style,
              strutStyle: strutStyle,
              textCapitalization: textCapitalization,
              autofocus: autofocus,
              readOnly: true,
              showCursor: showCursor,
              obscureText: obscureText,
              autocorrect: autocorrect,
              expands: expands,
              onChanged: (text) => field.didChange(state.parse(text)),
              onEditingComplete: onEditingComplete,
              onSubmitted: (text) => onSubmitted?.call(state.parse(text)),
              inputFormatters: inputFormatters,
              enabled: enabled,
              cursorWidth: cursorWidth,
              cursorRadius: cursorRadius,
              cursorColor: cursorColor,
              scrollPadding: scrollPadding,
              keyboardAppearance: keyboardAppearance,
              enableInteractiveSelection: enableInteractiveSelection,
              buildCounter: buildCounter,
            );
          },
        );

  final FormNumFieldPicker picker;

  final TextEditingController? controller;
  final FocusNode? focusNode;
  final bool readOnly;
  final void Function(num? value)? onChanged;

  @override
  _NumTextFieldState createState() => _NumTextFieldState();
}

class _NumTextFieldState extends FormFieldState<num> {
  TextEditingController? _controller;
  FocusNode? _focusNode;
  bool isShowingDialog = false;
  bool hadFocus = false;

  @override
  _NumTextField get widget => super.widget as _NumTextField;

  TextEditingController? get _effectiveController =>
      widget.controller ?? _controller;
  FocusNode? get _effectiveFocusNode => widget.focusNode ?? _focusNode;

  bool get hasFocus => _effectiveFocusNode?.hasFocus ?? false;
  bool get hasText => _effectiveController?.text.isNotEmpty ?? false;

  @override
  void initState() {
    super.initState();
    if (widget.controller == null) {
      _controller = TextEditingController(text: format(widget.initialValue));
      _controller?.addListener(_handleControllerChanged);
    }
    if (value == null) {
      setValue(parse(_effectiveController?.text ?? "") ?? widget.initialValue);
    }
    if (widget.focusNode == null) {
      _focusNode = FocusNode();
      _focusNode?.addListener(_handleFocusChanged);
    }
    widget.controller?.addListener(_handleControllerChanged);
    widget.focusNode?.addListener(_handleFocusChanged);
  }

  @override
  void didUpdateWidget(_NumTextField oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (widget.controller != oldWidget.controller) {
      oldWidget.controller?.removeListener(_handleControllerChanged);
      widget.controller?.addListener(_handleControllerChanged);

      if (oldWidget.controller != null && widget.controller == null) {
        _controller =
            TextEditingController.fromValue(oldWidget.controller?.value);
        _controller?.addListener(_handleControllerChanged);
      }
      if (widget.controller != null) {
        setValue(parse(widget.controller?.text ?? ""));
        if (oldWidget.controller == null) {
          _controller?.dispose();
          _controller = null;
        }
      }
    }
    if (widget.focusNode != oldWidget.focusNode) {
      oldWidget.focusNode?.removeListener(_handleFocusChanged);
      widget.focusNode?.addListener(_handleFocusChanged);

      if (oldWidget.focusNode != null && widget.focusNode == null) {
        _focusNode = FocusNode();
        _focusNode?.addListener(_handleFocusChanged);
      }
      if (widget.focusNode != null && oldWidget.focusNode == null) {
        // Release the focus node since it wont be used
        _focusNode?.dispose();
        _focusNode = null;
      }
    }
  }

  @override
  void didChange(num? value) {
    widget.onChanged?.call(value);
    super.didChange(value);
  }

  @override
  void dispose() {
    _controller?.dispose();
    _focusNode?.dispose();
    widget.controller?.removeListener(_handleControllerChanged);
    widget.focusNode?.removeListener(_handleFocusChanged);
    super.dispose();
  }

  @override
  void reset() {
    super.reset();
    _effectiveController?.text = format(widget.initialValue) ?? "";
    didChange(widget.initialValue);
  }

  void _handleControllerChanged() {
    if (_effectiveController?.text != format(value))
      didChange(parse(_effectiveController?.text));
  }

  String? format(num? value) => value?.toString();
  num? parse(String? text) {
    try {
      return text.isEmpty ? null : _parseLoose(text!);
    } catch (e) {
      return null;
    }
  }

  num? _parseLoose(String text) {
    final i = int.tryParse(text);
    if (i != null) {
      return i;
    }
    return double.tryParse(text);
  }

  Future<void> requestUpdate() async {
    if (widget.readOnly) {
      return;
    }
    if (!isShowingDialog) {
      isShowingDialog = true;
      final newValue = await widget.picker.build(context, value);
      isShowingDialog = false;
      if (newValue != null) {
        _effectiveController?.text = format(newValue) ?? "";
      }
    }
  }

  void _handleFocusChanged() {
    if (hasFocus && !hadFocus) {
      hadFocus = hasFocus;
      _hideKeyboard();
      requestUpdate();
    } else {
      hadFocus = hasFocus;
    }
  }

  void _hideKeyboard() {
    Future.microtask(() => FocusScope.of(context).requestFocus(FocusNode()));
  }

  Future<void> clear() async {
    if (widget.readOnly) {
      return;
    }
    _hideKeyboard();
    WidgetsBinding.instance.scheduleFrameCallback((_) {
      setState(() => _effectiveController?.clear());
    });
  }
}

class FormNumFieldPicker {
  const FormNumFieldPicker({
    this.defaultValue,
    this.begin = 0,
    this.interval = 1,
    this.end = 100,
    this.suffix = "",
    this.backgroundColor,
    this.color,
    this.confirmText = "Confirm",
    this.cancelText = "Cancel",
  })  : assert(begin < end, "[begin] must be less than [end]."),
        assert(interval > 0, "[interval] must be greater than 0.");

  final num? defaultValue;
  final num begin;
  final num interval;
  final num end;
  final String suffix;
  final Color? backgroundColor;
  final Color? color;
  final String confirmText;
  final String cancelText;

  Future<num?> build(BuildContext context, num? currentValue) async {
    final length = (end - begin) ~/ interval;
    num? res;
    await Picker(
      height: 240,
      backgroundColor: backgroundColor ?? Theme.of(context).colorScheme.surface,
      containerColor: backgroundColor ?? Theme.of(context).colorScheme.surface,
      headerColor: backgroundColor ?? Theme.of(context).colorScheme.surface,
      textStyle:
          TextStyle(color: color ?? Theme.of(context).colorScheme.onSurface),
      confirmText: confirmText,
      cancelText: cancelText,
      selecteds: [
        ((currentValue ?? defaultValue ?? 0) - begin) ~/ interval,
      ],
      adapter: PickerDataAdapter<num>(
        data: [
          ...List.generate(length, (m) {
            final val = (begin + m) * interval;
            return PickerItem<num>(
              text: Text(
                "$val$suffix",
                style: TextStyle(
                  color: color ?? Theme.of(context).colorScheme.onSurface,
                ),
              ),
              value: val,
            );
          }),
        ],
      ),
      changeToFirst: true,
      hideHeader: false,
      onConfirm: (Picker picker, List<int> value) {
        res = (begin + value[0]) * interval;
      },
    ).showModal(context);
    return res;
  }
}
